import re
from typing import Tuple, Dict, Optional

from gazpacho import Soup
from returns.pipeline import flow

from models.Car import CarDate, CarFull, Country, FuelType, BodyType, Drivetrain
from scrapers.utils import parse_price, find_one, find_many

DOMAIN = "https://lietotiauto.mollerauto.lv"


def scrape_moller_car_detail(html: str) -> CarFull:
    soup = Soup(html)

    def get_row_tuple(row: Soup) -> Tuple[str, str]:
        def extract_text(s: Soup) -> str:
            return s.text

        name: str = flow(find_one("th")(row), lambda _: _.bind(extract_text))
        value: str = flow(find_one("td")(row), lambda _: _.bind(extract_text))
        return name, value

    def get_table_dict(which_table: int) -> Dict[str, str]:
        return flow(
            soup,
            find_many("table"),
            lambda _: _.bind(lambda v: v[which_table]),
            find_many("tr"),
            lambda _: _.bind(lambda rows: tuple(get_row_tuple(row) for row in rows)),
            dict
        )

    info = get_table_dict(0)
    details = get_table_dict(1)
    imageSrc = flow(find_one("a", {"class": "fancybox"})(soup), lambda _: _.bind(lambda v: v.attrs["href"]))
    url = flow(
        find_one("meta", {"property": "og:url"})(soup),
        lambda _: _.map(lambda v: v.attrs["content"])
    ).value_or("")

    return CarFull(
        url=url,
        summary=find_one("h1")(soup).map(lambda _: _.text).value_or(""),
        transmission=details.get('Pārnesumkārba:'),
        hp=details.get('Jauda:'),
        type=details.get('Virsbūves tips:'),
        body=parse_body(details.get('Virsbūves tips:')),
        date=parse_date(details.get('Pirmā reģistrācija:')),
        color=details.get('Krāsa:'),
        engineSize=parse_int(details.get('Dzinēja tilpums:')),
        mileage=parse_int(details.get('Nobraukums:')),
        fuelType=parse_fuel_type(details.get('Degviela:')),
        hasWarranty=parse_warranty(details.get('Garantija:')),
        doors=details.get('Durvju skaits:'),
        vin=details.get('Virsbūves (VIN) numurs:'),
        registrationNo=details.get('Reģistrācijas numurs:'),
        techInspDate=parse_date(details.get('Tehniskā skate:')),
        drivetrain=parse_drivetrain(details.get('Piedziņa:')),
        price=parse_price(info['Cena:']),
        dealer="moller-auto",
        country=parse_country(info.get('Valsts:')),
        previewImgSrc="https://lietotiauto.mollerauto.lv" + imageSrc
    )


def parse_country(country: Optional[str]) -> Optional[Country]:
    if country == "Latvija":
        return "lv"
    if country == "Lietuva":
        return "lt"
    if country == "Igaunija":
        return "ee"
    return None


def parse_date(date: Optional[str]) -> Optional[CarDate]:
    if date is None:
        return None

    d = date.split("/")
    return CarDate(d[0], d[1])


def parse_int(mileage: Optional[str]) -> Optional[int]:
    if mileage is None:
        return None

    return int(re.findall(r'\d+', mileage.replace(" ", ""))[0])


def parse_fuel_type(fuelType: Optional[str]) -> Optional[FuelType]:
    if fuelType == "benzīns":
        return "petrol"
    if fuelType == "dīzelis":
        return "diesel"
    return None


def parse_body(body: Optional[str]) -> Optional[BodyType]:
    if body == "sedans":
        return "sedan"
    if body == "universāls":
        return "wagon"
    if body == "pikaps":
        return "pickup"
    return None


def parse_warranty(warranty: Optional[str]) -> Optional[bool]:
    if warranty == "Ražotāja garantija":
        return True
    return None


def parse_drivetrain(drivetrain: Optional[str]) -> Optional[Drivetrain]:
    if drivetrain == "pilnpiedziņa":
        return "awd"
    if drivetrain == "priekšējā piedziņa":
        return "fwd"
    if drivetrain == "aizmugures piedziņa":
        return "rwd"
    return None



import unittest
import os

from models.Car import CarFull, CarDate
from scrapers.details.inchcape import scrape_inchcape_car_detail

THIS_FOLDER = os.path.dirname(os.path.abspath(__file__))


class UtilsTestCase(unittest.TestCase):
    def test_bmw_x6m(self) -> None:
        with open(os.path.join(THIS_FOLDER, "inchcape/inchcape_bmw_x6m.html")) as car:
            self.assertEqual(CarFull(
                url='https://certified.inchcape.lv/auto/bmw/bmw-x6-m-2016-44-423-kw-f86-30393-02112021100237',
                previewImgSrc='https://certified.inchcape.lv/assets/images/auto/34286/61a0aedb3807b.jpg',
                summary='BMW X6 M 2016 4.4   423 kW F86, 133543 km',
                date=CarDate(month='06', year='2016'),
                type='SUV',
                transmission='Automātiskā',
                hp='423 kW',
                price=54900,
                vin='WBSKW810100G93368',
                registrationNo=None,
                mileage=133543,
                engineSize=None,
                techInspDate=None,
                fuelType='petrol',
                body="suv",
                drivetrain='awd',
                color=None,
                hasWarranty=True,
                doors=None,
                country='lv',
                dealer='inchcape'
            ), scrape_inchcape_car_detail(car.read()))

    def test_ford_focus(self) -> None:
        with open(os.path.join(THIS_FOLDER, "inchcape/inchcape_ford_focus.html")) as car:
            self.assertEqual(CarFull(
                url='https://certified.inchcape.lv/auto/ford/ford-focus-2016-10-fwd-92-kw-ford1-14092021144651',
                previewImgSrc='https://certified.inchcape.lv/assets/images/auto/34066/61781015e92b9.jpg',
                summary='Ford Focus 2016 1.0  FWD 92 kW, 160247 km',
                date=CarDate(month='11', year='2016'),
                type='Hečbeks',
                transmission='Automātiskā',
                hp='92 kW',
                price=9990,
                vin='WF05XXGCC5GY15636',
                registrationNo=None,
                mileage=160247,
                engineSize=None,
                techInspDate=None,
                fuelType='petrol',
                body="hatchback",
                drivetrain='fwd',
                color=None,
                hasWarranty=True,
                doors=None,
                country='lv',
                dealer='inchcape'
            ), scrape_inchcape_car_detail(car.read()))

    def test_bmw_530(self) -> None:
        with open(os.path.join(THIS_FOLDER, "inchcape/inchcape_bmw_530.html")) as car:
            self.assertEqual(CarFull(
                url='https://certified.inchcape.lv/auto/bmw/bmw-530-2019-20-rwd-135-kw-g30-iperformance-30393-05102021164152',
                previewImgSrc='https://certified.inchcape.lv/assets/images/auto/33944/61603d224f07d.jpg',
                summary='BMW 530 2019 2.0  RWD 135 kW G30 , Iperformance, 47522 km',
                date=CarDate(month='07', year='2019'),
                type='Sedans',
                transmission='Automātiskā',
                hp='135 kW',
                price=46800,
                vin='WBAJA910X0GF99601',
                registrationNo=None,
                mileage=47522,
                engineSize=None,
                techInspDate=None,
                fuelType="hybrid",
                body="sedan",
                drivetrain='rwd',
                color=None,
                hasWarranty=True,
                doors=None,
                country='lv',
                dealer='inchcape'
            ), scrape_inchcape_car_detail(car.read()))



if __name__ == '__main__':
    unittest.main()



import asyncio
import functools
from typing import TypeVar, Tuple, Dict, Callable

from mypy_extensions import VarArg, KwArg
from returns.future import future_safe, FutureResultE

T = TypeVar('T')
T1 = TypeVar('T1')
R = TypeVar('R')


def sync_to_async(fn: Callable[..., R]) -> Callable[[VarArg(Tuple[T, ...]), KwArg(Dict[str, T1])], FutureResultE[R]]:
    @future_safe
    async def wrapper(*args: Tuple[T, ...], **kwargs: Dict[str, T1]) -> R:
        loop = asyncio.get_event_loop()
        p_func = functools.partial(fn, *args, **kwargs)
        return await loop.run_in_executor(None, p_func)

    return wrapper



from typing import Iterable

import requests
from gazpacho import Soup
from returns.maybe import Maybe
from returns.pipeline import flow

from scrapers.utils import find_one, find_many

DOMAIN = "https://lietotiauto.mollerauto.lv"


def fetch_moller_urls(page: int = 1) -> Iterable[str]:
    print("fetch_moller_urls")
    r = requests.post(
        url=f"{DOMAIN}/lv/usedcars/search",
        data={"ajaxsearch": 1, "search_drivetrain": 10003016, "page": page},
    )
    soup = Soup(r.text.encode().decode("unicode-escape").replace("\/", "/"))

    return flow(
        soup,
        find_many("div", {"class": "vehicle"}),
        lambda _: _.bind(lambda _: _),
        lambda _: [flow(
            car,
            find_one("div", {"class": "vehiclesummary"}),
            lambda _: _.bind(find_one("a")),
            lambda _: _.bind(lambda _: f"{DOMAIN}{get_url(_)}"),
        ) for car in _],
    )


def get_url(summary: Soup) -> str:
    return Maybe.from_optional(summary.attrs).map(lambda v: v["href"]).value_or("")



import asyncio
import time
from itertools import islice, takewhile, repeat, chain
from typing import List, Iterable, Sequence, TypeVar, Callable, Tuple

import httpx
from returns.functions import tap
from returns.pipeline import flow
from returns.converters import flatten
from returns.future import future_safe, FutureResultE
from returns.io import IOResultE
from returns.iterables import Fold
from returns.pointfree import map_

from models.Car import CarFull
from scrapers.details.brcAuto import scrape_brc_auto_car_detail
from scrapers.details.inchcape import scrape_inchcape_car_detail
from scrapers.details.mollerAuto import scrape_moller_car_detail
from scrapers.list.brcAuto import fetch_brc_auto_urls
from scrapers.list.mollerAuto import fetch_moller_urls
from scrapers.list.inchcape import fetch_inchcape_urls

from utils.sync_to_async import sync_to_async

ListScraper = Callable[[int], Iterable[str]]
CarScraper = Callable[[str], CarFull]
CarHtmlFutures = Iterable[FutureResultE[str]]
CarFullFutures = Iterable[FutureResultE[CarFull]]
T = TypeVar('T')
_FirstType = TypeVar('_FirstType')
_SecondType = TypeVar('_SecondType')


async def scrape_all() -> Iterable[IOResultE[CarFull]]:
    return await scrape_specific([
        (fetch_moller_urls, scrape_moller_car_detail),
        (fetch_inchcape_urls, scrape_inchcape_car_detail),
        (fetch_brc_auto_urls, scrape_brc_auto_car_detail),
    ])


async def scrape_specific(scrapers: Iterable[Tuple[ListScraper, CarScraper]]) -> Iterable[IOResultE[CarFull]]:
    car_futures: Sequence[FutureResultE[CarFullFutures]] = [
        fetch_car_pages(sync_to_async(list_scraper), car_scraper)
        for (list_scraper, car_scraper) in scrapers
    ]

    cars: Sequence[IOResultE[Tuple[CarFullFutures, ...]]] = await asyncio.gather(*car_futures)

    return flow(
        Fold.collect(cars, IOResultE.from_value(())),
        lambda _: _.bind(lambda _: _),
        lambda _: flatten_iter(_),
    )


def fetch_car_pages(
        get_urls: Callable[..., FutureResultE[Iterable[str]]],
        car_parser: Callable[[str], CarFull]
) -> FutureResultE[CarFullFutures]:
    future_cars: FutureResultE[Iterable[str]] = get_urls()

    car_htmls: FutureResultE[CarHtmlFutures] = flow(
        future_cars,
        map_(fetch_cars_in_batches),
        flatten
    )

    return flow(
        car_htmls,
        lambda _: _.map(lambda _: [car.map(car_parser) for car in _]),
    )


def fetch_cars_in_batches(res: Iterable[str]) -> FutureResultE[CarHtmlFutures]:
    url_batches = split_every(10, res)
    futures_of_batched_car_details_futures = tuple([fetch_car_htmls(url_batch) for url_batch in url_batches])

    future_of_batched_car_details_futures = flow(
        futures_of_batched_car_details_futures,
        lambda _: Fold.collect(_, FutureResultE.from_value(()))
    )

    future_of_car_details_futures: FutureResultE[CarHtmlFutures] = flow(
        future_of_batched_car_details_futures,
        map_(flatten_iter)
    )

    return future_of_car_details_futures


@future_safe
async def fetch_car_htmls(urls: Iterable[str]) -> CarHtmlFutures:
    return iter(await asyncio.gather(
        *[fetch_car_html(u) for u in urls]
    ))


@future_safe
async def fetch_car_html(url: str) -> str:
    async with httpx.AsyncClient(timeout=15) as client:
        response = await client.get(url)
        response.raise_for_status()
        return response.text


def split_every(n: int, iterable: Iterable[T]) -> Iterable[Iterable[T]]:
    iterator = iter(iterable)
    return takewhile(bool, (list(islice(iterator, n)) for _ in repeat(None)))


def flatten_iter(v: Iterable[Iterable[T]]) -> Iterable[T]:
    return chain(*v)
